@@ -2,28 +2,34 @@ import asyncio
 import functools
 import os
 import signal
+import socket
+import ssl
 import sys
 import time
+import typing
 from email.utils import formatdate
 
 import click
 
 from uvicorn.config import (
     HTTP_PROTOCOLS,
+    INTERFACES,
     LIFESPAN,
     LOG_LEVELS,
     LOOP_SETUPS,
+    SSL_PROTOCOL_VERSION,
     WS_PROTOCOLS,
     Config,
     get_logger,
 )
-from uvicorn.reloaders.statreload import StatReload
+from uvicorn.supervisors import Multiprocess, StatReload
 
 LEVEL_CHOICES = click.Choice(LOG_LEVELS.keys())
 HTTP_CHOICES = click.Choice(HTTP_PROTOCOLS.keys())
 WS_CHOICES = click.Choice(WS_PROTOCOLS.keys())
 LIFESPAN_CHOICES = click.Choice(LIFESPAN.keys())
 LOOP_CHOICES = click.Choice(LOOP_SETUPS.keys())
+INTERFACE_CHOICES = click.Choice(INTERFACES)
 
 HANDLED_SIGNALS = (
     signal.SIGINT,  # Unix signal 2. Sent by Ctrl+C.
@@ -51,6 +57,22 @@ HANDLED_SIGNALS = (
 @click.option(
     "--fd", type=int, default=None, help="Bind to socket from this file descriptor."
 )
+@click.option(
+    "--debug", is_flag=True, default=False, help="Enable debug mode.", hidden=True
+)
+@click.option("--reload", is_flag=True, default=False, help="Enable auto-reload.")
+@click.option(
+    "--reload-dir",
+    "reload_dirs",
+    multiple=True,
+    help="Set reload directories explicitly, instead of using 'sys.path'.",
+)
+@click.option(
+    "--workers",
+    default=1,
+    type=int,
+    help="Number of worker processes. Not valid with --reload.",
+)
 @click.option(
     "--loop",
     type=LOOP_CHOICES,
@@ -80,12 +102,12 @@ HANDLED_SIGNALS = (
     show_default=True,
 )
 @click.option(
-    "--wsgi",
-    is_flag=True,
-    default=False,
-    help="Use WSGI as the application interface, instead of ASGI.",
+    "--interface",
+    type=INTERFACE_CHOICES,
+    default="auto",
+    help="Select ASGI3, ASGI2, or WSGI as the application interface.",
+    show_default=True,
 )
-@click.option("--debug", is_flag=True, default=False, help="Enable debug mode.")
 @click.option(
     "--log-level",
     type=LEVEL_CHOICES,
@@ -127,6 +149,50 @@ HANDLED_SIGNALS = (
     help="Close Keep-Alive connections if no new data is received within this timeout.",
     show_default=True,
 )
+@click.option(
+    "--ssl-keyfile", type=str, default=None, help="SSL key file", show_default=True
+)
+@click.option(
+    "--ssl-certfile",
+    type=str,
+    default=None,
+    help="SSL certificate file",
+    show_default=True,
+)
+@click.option(
+    "--ssl-version",
+    type=int,
+    default=SSL_PROTOCOL_VERSION,
+    help="SSL version to use (see stdlib ssl module's)",
+    show_default=True,
+)
+@click.option(
+    "--ssl-cert-reqs",
+    type=int,
+    default=ssl.CERT_NONE,
+    help="Whether client certificate is required (see stdlib ssl module's)",
+    show_default=True,
+)
+@click.option(
+    "--ssl-ca-certs",
+    type=str,
+    default=None,
+    help="CA certificates file",
+    show_default=True,
+)
+@click.option(
+    "--ssl-ciphers",
+    type=str,
+    default="TLSv1",
+    help="Ciphers to use (see stdlib ssl module's)",
+    show_default=True,
+)
+@click.option(
+    "--header",
+    "headers",
+    multiple=True,
+    help="Specify custom default HTTP response headers as a Name:Value pair",
+)
 def main(
     app,
     host: str,
@@ -137,8 +203,11 @@ def main(
     http: str,
     ws: str,
     lifespan: str,
-    wsgi: bool,
+    interface: str,
     debug: bool,
+    reload: bool,
+    reload_dirs: typing.List[str],
+    workers: int,
     log_level: str,
     no_access_log: bool,
     proxy_headers: bool,
@@ -146,6 +215,13 @@ def main(
     limit_concurrency: int,
     limit_max_requests: int,
     timeout_keep_alive: int,
+    ssl_keyfile: str,
+    ssl_certfile: str,
+    ssl_version: int,
+    ssl_cert_reqs: int,
+    ssl_ca_certs: str,
+    ssl_ciphers: str,
+    headers: typing.List[str],
 ):
     sys.path.insert(0, ".")
 
@@ -161,27 +237,46 @@ def main(
         "lifespan": lifespan,
         "log_level": log_level,
         "access_log": not no_access_log,
-        "wsgi": wsgi,
+        "interface": interface,
         "debug": debug,
+        "reload": reload,
+        "reload_dirs": reload_dirs if reload_dirs else None,
+        "workers": workers,
         "proxy_headers": proxy_headers,
         "root_path": root_path,
         "limit_concurrency": limit_concurrency,
         "limit_max_requests": limit_max_requests,
         "timeout_keep_alive": timeout_keep_alive,
+        "ssl_keyfile": ssl_keyfile,
+        "ssl_certfile": ssl_certfile,
+        "ssl_version": ssl_version,
+        "ssl_cert_reqs": ssl_cert_reqs,
+        "ssl_ca_certs": ssl_ca_certs,
+        "ssl_ciphers": ssl_ciphers,
+        "headers": list([header.split(":") for header in headers]),
     }
-
-    if debug:
-        logger = get_logger(log_level)
-        reloader = StatReload(logger)
-        reloader.run(run, kwargs)
-    else:
-        run(**kwargs)
+    run(**kwargs)
 
 
 def run(app, **kwargs):
     config = Config(app, **kwargs)
     server = Server(config=config)
-    server.run()
+
+    if config.reload and not isinstance(app, str):
+        config.logger_instance.warn(
+            "auto-reload only works when app is passed as an import string."
+        )
+
+    if isinstance(app, str) and (config.debug or config.reload):
+        sock = config.bind_socket()
+        supervisor = StatReload(config)
+        supervisor.run(server.run, sockets=[sock])
+    elif config.workers > 1:
+        sock = config.bind_socket()
+        supervisor = Multiprocess(config)
+        supervisor.run(server.run, sockets=[sock])
+    else:
+        server.run()
 
 
 class ServerState:
@@ -204,68 +299,86 @@ class Server:
         self.started = False
         self.should_exit = False
         self.force_exit = False
+        self.last_notified = 0
 
-    def run(self):
+    def run(self, sockets=None, shutdown_servers=True):
+        self.config.setup_event_loop()
+        loop = asyncio.get_event_loop()
+        loop.run_until_complete(self.serve(sockets=sockets))
+
+    async def serve(self, sockets=None, shutdown_servers=True):
         process_id = os.getpid()
 
         config = self.config
         if not config.loaded:
             config.load()
 
-        self.loop = config.loop_instance
         self.logger = config.logger_instance
         self.lifespan = config.lifespan_class(config)
 
         self.install_signal_handlers()
 
         self.logger.info("Started server process [{}]".format(process_id))
-        self.loop.run_until_complete(self.startup())
-        self.loop.run_until_complete(self.main_loop())
-        self.loop.run_until_complete(self.shutdown())
-        self.loop.stop()
+        await self.startup(sockets=sockets)
+        if self.should_exit:
+            return
+        await self.main_loop()
+        await self.shutdown(shutdown_servers=shutdown_servers)
         self.logger.info("Finished server process [{}]".format(process_id))
 
-    async def startup(self):
+    async def startup(self, sockets=None):
         config = self.config
 
         await self.lifespan.startup()
+        if self.lifespan.should_exit:
+            self.should_exit = True
+            return
 
         create_protocol = functools.partial(
             config.http_protocol_class, config=config, server_state=self.server_state
         )
 
-        if config.sockets is not None:
+        loop = asyncio.get_event_loop()
+
+        if sockets is not None:
             # Explicitly passed a list of open sockets.
             # We use this when the server is run from a Gunicorn worker.
             self.servers = []
-            for socket in config.sockets:
-                server = await self.loop.create_server(create_protocol, sock=socket)
+            for sock in sockets:
+                server = await loop.create_server(
+                    create_protocol, sock=sock, ssl=config.ssl
+                )
                 self.servers.append(server)
 
         elif config.fd is not None:
             # Use an existing socket, from a file descriptor.
             sock = socket.fromfd(config.fd, socket.AF_UNIX, socket.SOCK_STREAM)
-            server = await self.loop.create_server(create_protocol, sock=sock)
+            server = await loop.create_server(
+                create_protocol, sock=sock, ssl=config.ssl
+            )
             message = "Uvicorn running on socket %s (Press CTRL+C to quit)"
             self.logger.info(message % str(sock.getsockname()))
             self.servers = [server]
 
         elif config.uds is not None:
             # Create a socket using UNIX domain socket.
-            server = await self.loop.create_unix_server(
-                create_protocol, path=config.uds
-            )
+            uds_perms = 0o666
+            if os.path.exists(config.uds):
+                uds_perms = os.stat(config.uds).st_mode
+            server = await loop.create_unix_server(create_protocol, path=config.uds)
+            os.chmod(config.uds, uds_perms)
             message = "Uvicorn running on unix socket %s (Press CTRL+C to quit)"
             self.logger.info(message % config.uds)
             self.servers = [server]
 
         else:
             # Standard case. Create a socket from a host/port pair.
-            server = await self.loop.create_server(
-                create_protocol, host=config.host, port=config.port
+            server = await loop.create_server(
+                create_protocol, host=config.host, port=config.port, ssl=config.ssl
             )
-            message = "Uvicorn running on http://%s:%d (Press CTRL+C to quit)"
-            self.logger.info(message % (config.host, config.port))
+            protocol_name = "https" if config.ssl else "http"
+            message = "Uvicorn running on %s://%s:%d (Press CTRL+C to quit)"
+            self.logger.info(message % (protocol_name, config.host, config.port))
             self.servers = [server]
 
         self.started = True
@@ -285,14 +398,14 @@ class Server:
             current_time = time.time()
             current_date = formatdate(current_time, usegmt=True).encode()
             self.server_state.default_headers = [
-                (b"server", b"uvicorn"),
-                (b"date", current_date),
-            ]
+                (b"date", current_date)
+            ] + self.config.encoded_headers
 
-        # Callback to `callback_notify` once every `timeout_notify` seconds.
-        if self.config.callback_notify is not None:
-            if counter % (10 * self.config.timeout_notify) == 0:
-                await self.config.callback_notify()
+            # Callback to `callback_notify` once every `timeout_notify` seconds.
+            if self.config.callback_notify is not None:
+                if current_time - self.last_notified > self.config.timeout_notify:
+                    self.last_notified = current_time
+                    await self.config.callback_notify()
 
         # Determine if we should exit.
         if self.should_exit:
@@ -301,11 +414,11 @@ class Server:
             return self.server_state.total_requests >= self.config.limit_max_requests
         return False
 
-    async def shutdown(self):
+    async def shutdown(self, shutdown_servers=True):
         self.logger.info("Shutting down")
 
         # Stop accepting new connections.
-        if not self.config.sockets:
+        if shutdown_servers:
             for server in self.servers:
                 server.close()
             for server in self.servers:
@@ -335,9 +448,11 @@ class Server:
             await self.lifespan.shutdown()
 
     def install_signal_handlers(self):
+        loop = asyncio.get_event_loop()
+
         try:
             for sig in HANDLED_SIGNALS:
-                self.loop.add_signal_handler(sig, self.handle_exit, sig, None)
+                loop.add_signal_handler(sig, self.handle_exit, sig, None)
         except NotImplementedError as exc:
             # Windows
             for sig in HANDLED_SIGNALS:
